/**
 * Please note that this code may not be specifically used by any of the
 * samples; however, it is provided as a template for a screen class should
 * you need one that provides more functionality than the basic screen.cpp/h.
 * The basic idea is that you would copy screen_template.cpp/h over to your
 * own source tree, rename it appropriately, and modify it to add whatever
 * functionality you would need.
 */
#include <sys/ioctl.h>
#include <sys/vt.h>
#include <linux/ps2/dev.h>
#include <linux/kd.h>
#include <stdio.h>
#include "screen.h"
#include "p2lassert.h"
#include "dmagif.h"


// 512 pages in buffer
void CScreen::Setup()
{	
	//uint32 uVCKv = 1;
	uint32 uGSPage = 0;
	//uint32 uMask = 0;		
	uint32 uMode = 0;	// 32bit
	//uint32 uVCKh, uDispX, uDispY;

	// Setup Vesa
	switch(auto_detect_video_mode())
	{
		case SCR_VESA:
			reset_gs_vesa();
			break;
		case SCR_NTSC:
			reset_gs_ntsc();
			break;
		case SCR_PAL:
			reset_gs_pal();
			break;
		default:
			exit(0);
	}
	
	/*if(uScrMode == 0)	// NTSC	(640x448, 32bit, field, interlace)
	{
		uVCKh = 4;		// PixelSize in VCK units on screen 
		uDispX = 0x27c;		// display posX in VCK units, for some reason 
		uDispY = 0x32;
		m_uWidth = 640;
		m_uHeight = 448;
		//GsResetGraph(0, GS_INTERLACE, GS_NTSC, GS_FIELD);	// read each line
	}
	else			// PAL	(512x512, 32bit, field, interlace)
	{
		uVCKh = 5;
		uDispX = 0x290;
		uDispY = 0x48;
		m_uWidth = 512;
		m_uHeight = 512;
		//GsResetGraph(0, GS_INTERLACE, GS_PAL, GS_FIELD);	// read each line
	}*/

	//FlushCache(0);
	*EE_GIF_MODE = 4;	// enable intermittent mode.

	// *GS_PMODE(m_pvGSRegisters) = PMODE_SET(0, 1, 1, 1, 1, 0, 0);
	// *GS_SMODE2(m_pvGSRegisters) = SMODE2_SET(0,0,0);
	DPUT_GS_DISPFB2(DISPFB_SET(uGSPage, m_uWidth/64, uMode, 0, 0));
	// *GS_DISPLAY2(m_pvGSRegisters) = DISPLAY_SET(uDispX, uDispY, uVCKh-1, uVCKv-1, m_uWidth*uVCKh-1, m_uHeight-1);
	// *GS_BGCOLOR(m_pvGSRegisters) = 0x0;

	SetBuffers();
}



void CScreen::ScreenSetup()
{
	//GsResetPath();
   //DmaReset(1);
	*EE_VIF1_ERR = 2;

	Setup();
}


void CScreen::Swap()
{
	VSync(0);              // Wait for vertical blank
	buffer = 1-buffer;
							
	uint32 uPage = (buffer == 0) ? m_uPtr1 : m_uPtr2;
	uint32 uMask = 0;		
	uint32 uMode = 0;	// 32bit

	uint32 uDispPage = (buffer == 0) ? m_uPtr2 : m_uPtr1;
	DPUT_GS_DISPFB2(DISPFB_SET(uDispPage, m_uWidth/64, uMode, 0, 0));

	// various data args
	const uint32 nloop = 13;
	const uint32 uOffsX = (2048-(m_uWidth>>1))*16;				// Now center in render will always be 2048, 2048
	const uint32 uOffsY = (2048-(m_uHeight>>1))*16;
	const uint32 uEndX = (2048+(m_uWidth>>1))*16;
	const uint32 uEndY = (2048+(m_uHeight>>1))*16;

	// gif codes data
	const uint64 uGifData[] _ALIGN16 =
	{
		GIFTAG(nloop, 1, 0, 0, 0, 1),
		GIFTAG_REGS1(AD),

		FRAME_SET(uPage,m_uWidth/64,uMode,uMask),
		FRAME_1,

		ZBUF_SET(m_zBufPtr, 1, 0),
		ZBUF_1,

		XYOFFSET_SET(uOffsX, uOffsY),
		XYOFFSET_1,

		SCISSOR_SET(0, m_uWidth-1, 0, m_uHeight-1),
		SCISSOR_1,

		PRMODECONT_SET(0x1),
		PRMODECONT,

		COLCLAMP_SET(0x1),
		COLCLAMP,
		
		DTHE_SET(0x0),
		DTHE,

		///////////////////////// This last part is added to clear the screen
		TEST_SET(0, 0, 0, 0, 0, 0, 1, 1),
		TEST_1,

		PRIM_SET(0x6, 0, 0, 0, 0, 0, 0, 0, 0),
		PRIM,
		
		RGBAQ_SET(0x10, 0x10, 0x20, 0x80, 0x3f800000),
		RGBAQ,

		XYZ2_SET(uOffsX, uOffsY, 0),
		XYZ2,

		XYZ2_SET(uEndX, uEndY, 0),
		XYZ2,
	
		TEST_SET(0, 0, 0, 0, 0, 0, 1, 3),
		TEST_1
	};

	// execute gif codes
	for(uint32 q=0; q<nloop+1; q++)
		DPUT_EE_GIF_FIFO(((uint128 *) uGifData)[q]);

	//GsSyncPath(0, 0);
}
  
uint32 CScreen::GetFirstFreeGsPage()
{
	return m_uRemBufPtr;
}

uint32 CScreen::GetDrawFptr()
{
	return (buffer == 0) ? m_uPtr1 : m_uPtr2;
}
	
uint32 CScreen::GetZptr()
{
	return m_zBufPtr;
}

uint32 CScreen::GetDrawBuff1()
{
	return m_uPtr1;
}

uint32 CScreen::GetDrawBuff2()
{
	return m_uPtr2;
}
	
void CScreen::SetBuffers()
{
	// locked at 32bit
	// 512 pages in GS buffer
	m_uPtr1=0;									
	m_uPtr2=(m_uWidth*m_uHeight)/2048;			
	m_zBufPtr=m_uPtr2+((m_uWidth*m_uHeight)/2048);
	m_uRemBufPtr=m_zBufPtr+((m_uWidth*m_uHeight)/2048);
	assert(((m_uWidth*m_uHeight) % 2048) == 0);
}


// PS2Linux stuff (libps2dev)
CScreen::CScreen()
{
	virtual_console_start();
	m_pOldScrInfo = (void *) new ps2_screeninfo;
	
	buffer = 0;
	m_inter = PS2_GS_NOINTERLACE;;
	
	m_fd_gs = open(PS2_DEV_GS, O_RDWR);
	if(m_fd_gs < 0)
	{
		perror("open(PS2_DEV_GS)");
		exit(0);
	}
	m_fd_event = open(PS2_DEV_EVENT, O_RDWR);
	if(m_fd_event < 0)
	{
		perror("open(PS2_DEV_EVENT)");
		close(m_fd_gs);
		exit(0);
	}

	ioctl(m_fd_gs, PS2IOC_GSCREENINFO, (ps2_screeninfo *) m_pOldScrInfo);
	ioctl(m_fd_event, PS2IOC_ENABLEEVENT, PS2EV_VBSTART);
}

CScreen::~CScreen()
{
	if (m_fd_gs==-1) return;
	ioctl(m_fd_gs, PS2IOC_SSCREENINFO, (ps2_screeninfo *) m_pOldScrInfo);
	if(m_fd_event >= 0) close(m_fd_event);
	if(m_fd_gs >= 0) close(m_fd_gs);
	delete ((ps2_screeninfo *) m_pOldScrInfo);
	virtual_console_stop();
}

void CScreen::reset_gs_vesa()
{	
	m_uWidth = SCR_WIDTH;
	m_uHeight = SCR_HEIGHT_VESA;
	m_uDepth = SCR_DEPTH;
	reset_gs(PS2_GS_NOINTERLACE, PS2_GS_VESA, 0, PS2_GS_640x480, PS2_GS_60Hz);
}

void CScreen::reset_gs_ntsc()
{	
	m_uWidth = SCR_WIDTH;
	m_uHeight = SCR_HEIGHT_NTSC;
	m_uDepth = SCR_DEPTH;
	reset_gs(PS2_GS_INTERLACE, PS2_GS_NTSC, PS2_GS_FIELD, 0/*PS2_GS_640x448*/, 0);
}

void CScreen::reset_gs_pal()
{	
	m_uWidth = SCR_WIDTH;
	m_uHeight = SCR_HEIGHT_PAL;
	m_uDepth = SCR_DEPTH;
	reset_gs(PS2_GS_INTERLACE, PS2_GS_PAL, PS2_GS_FIELD, 0/*PS2_GS_640x512*/, 0);
}

void CScreen::reset_gs(int inter, int omode, int ffmode, int resolution, int refreshRate)
{	
	//ps2_gs_reset(0, PS2_GS_NOINTERLACE, PS2_GS_VESA, PS2_GS_FRAME, PS2_GS_640x480, PS2_GS_60Hz);
	// args
	m_inter = inter;
	
	// setup code
	int ret;
	__u64 csr;
		
	//struct ps2_screeninfo sinfo;
	struct ps2_crtmode crtmode;
	struct ps2_display display2;
	struct ps2_pmode pmode;
	int width, height;
	int interlaceFormat;
	
	// full reset GS
	/*ret = ioctl(m_fd_gs, PS2IOC_GSRESET, PS2_GSRESET_FULL);
	if (ret < 0) {
		//perror("ps2_gs_reset, ioctl PS2_GSRESET_FULL");
		exit(0);
	}*/
		
	// set pmode, display off
	ret = ioctl(m_fd_gs, PS2IOC_GPMODE, &pmode);
	pmode.sw = 0; // ch1: off, ch2: off
	ret = ioctl(m_fd_gs, PS2IOC_SPMODE, &pmode);
		
	interlaceFormat = (omode == PS2_GS_NTSC || omode == PS2_GS_PAL ||
					   (omode == PS2_GS_DTV && resolution == PS2_GS_1080I));

	// get GS version
	int reg = PS2_GSSREG_CSR;
	struct ps2_gssreg r;
	r.reg = reg;
	ioctl(m_fd_gs, PS2IOC_GGSSREG, &r/*eg*/);
	csr = r.val;
	//csr = __ps2_gs_get_sreg(PS2_GSSREG_CSR);
	
	
	width = 640;
	height = 480;
		
	switch (omode)
	{
	case PS2_GS_NTSC:
	case PS2_GS_PAL:
		width = 640;
		height = (omode == PS2_GS_NTSC) ? 448 : 512;
		break;
			
	case PS2_GS_DTV:
		switch (resolution) {
		case PS2_GS_480P: width = 720; height = 480; break;
		case PS2_GS_1080I: width = 1920; height = 1080; break;
		case PS2_GS_720P: width = 1024; height = 720; break;
		}
		break;
			
	case PS2_GS_VESA:
		switch (resolution) {
		case PS2_GS_640x480: width = 640; height = 480; break;
		case PS2_GS_800x600: width = 800; height = 600; break;
		case PS2_GS_1024x768: width = 1024; height = 768; break;
		case PS2_GS_1280x1024: width = 1280; height = 1024; break;
		}
	}
		
	if (interlaceFormat && (m_inter == PS2_GS_NOINTERLACE || ffmode == PS2_GS_FRAME)) {
		// fake NI or I(frame)
		height /= 2;
	}
		
	//ps2_gs_set_center(2048, 2048);
	//__ps2_gs_set_screen_size(width, height);
	
	crtmode.mode = omode;
		
	switch (omode)
	{
	case PS2_GS_NTSC:
	case PS2_GS_PAL:
	case PS2_GS_DTV:
		crtmode.res = m_inter | ffmode;
		break;
			
	case PS2_GS_VESA:
		crtmode.res = resolution | refreshRate;
		break;
			
	default:
		//return -1;
		exit(0);
	}
		
	//ret = ioctl(gp->fd_gs, PS2IOC_SSCREENINFO, &sinfo);
	ret = ioctl(m_fd_gs, PS2IOC_SCRTMODE, &crtmode);
	if (ret < 0) {
		//perror("ps2_gs_reset, PS2IOC_SSCREENINFO");
		exit(0);
	}
		
	// display
	display2.ch = 1; // use region 2
	display2.w = width;
	display2.h = height;
	display2.dx = 0;
	display2.dy = 0;
	if (interlaceFormat && (m_inter == PS2_GS_INTERLACE && ffmode == PS2_GS_FRAME))
		display2.h *= 2;
		
	ret = ioctl(m_fd_gs, PS2IOC_SDISPLAY, &display2);
	if (ret < 0) {
		//perror("ps2_gs_reset, PS2IOC_SDISPLAY");
		exit(0);
	}

	// set pmode, display on
	ret = ioctl(m_fd_gs, PS2IOC_GPMODE, &pmode);
	pmode.sw = 2; // ch1: off, ch2: off
	ret = ioctl(m_fd_gs, PS2IOC_SPMODE, &pmode);
}

void CScreen::VSync(int mode)
{
	//ps2_gs_sync_v(mode);
	struct ps2_gssreg reg;
	
	ioctl(m_fd_event, PS2IOC_GETEVENT, PS2EV_VBSTART);
	ioctl(m_fd_event, PS2IOC_WAITEVENT, PS2EV_VBSTART);

	//int interlace_mode = 1;
	
	if (m_inter == PS2_GS_INTERLACE) {
		reg.reg = PS2_GSSREG_CSR;
		if (ioctl(m_fd_gs, PS2IOC_GGSSREG, &reg) < 0) {
			//perror("ps2_gs_sync_v, ioctl");
			exit(0);
		}
		
		// int iRetVal = ((ps2_gssreg_csr *)&reg.val)->FIELD;
	}
	
	// int iRetVal = 1; // odd
}

void CScreen::virtual_console_start()
{
	char fname[32];
	int mode;
	int new_vc;
	int err;
	struct vt_stat vts;

	m_console_fd = open( "/dev/console", O_RDWR );

	if (m_console_fd<0)
	{
		perror("open /dev/console");
		exit(-1);
	}

	err=ioctl(m_console_fd, VT_OPENQRY, &new_vc);
	if (err==-1)
	{
		perror("ioctl(VT_OPENQRY)");
		exit(-1);
	}
	if (new_vc<=0)
	{
		perror("ioctl(VT_OPENQRY) (wrong new_vc returned)");
		exit(-1);
	}

	sprintf(fname, "/dev/tty%d", new_vc);

	setpgid(0,getppid());
	setsid();

	m_tty_fd = open(fname, O_RDWR);
	if (m_tty_fd<0)
	{
		perror(fname);
		exit(-1);
	}

	ioctl(m_tty_fd, VT_GETSTATE, &vts);
	m_startup_vc=vts.v_active;

	ioctl(m_tty_fd, VT_ACTIVATE, new_vc);

	while (ioctl(m_tty_fd, VT_WAITACTIVE, new_vc) < 0)
	{
		if ((errno != EAGAIN) && (errno != EINTR))
		{
			perror("ioctl(VT_WAITACTIVE)");
			exit(-1);
		}
		usleep(150000);
	}


	// clear the console
	write(m_tty_fd, "\e[H\e[J", 6);


	mode = KD_GRAPHICS;
	printf("Switching to graphics\n");
	ioctl( m_tty_fd, KDSETMODE, mode );
}

void CScreen::virtual_console_stop()
{
	int mode;

	mode = KD_TEXT;
	printf("Deleting new console\n");
	ioctl( m_tty_fd, KDSETMODE, mode );

	ioctl(m_tty_fd, VT_ACTIVATE, m_startup_vc);

	close( m_tty_fd );
	close( m_console_fd );
}

uint32 CScreen::auto_detect_video_mode()
{
	int err;
	struct ps2_screeninfo screeninfo;

	err=ioctl(m_fd_gs, PS2IOC_GSCREENINFO, &screeninfo);
	if (err==-1)
	{
		perror("ioctl(PS2IOC_GSCREENINFO)");
		exit(-1);
	}

	switch (screeninfo.mode)
	{
		case 0:
			printf("Autodetected VESA mode\n");
			return m_uMode=SCR_VESA;
		case 2:
			printf("Autodetected NTSC mode\n");
			return m_uMode=SCR_NTSC;
		case 3:
			printf("Autodetected PAL mode\n");
			return m_uMode=SCR_PAL;
	}

	fprintf(stderr, "Could not autodetect display mode. Using default (VESA)\n");
	return m_uMode=SCR_VESA;
}

