#include <stdio.h>
#include <sps2lib.h>

#include "packet.h"

/*
 * This small sample allocates an uncached memory area and then
 * maps a cached equivalent of it.
 */
int main() {
  int iSPS2Device;
  sps2Memory_t *pUncached, *pCached;
  void *pvEERegisters, *pvVUMemory;
  unsigned long ulDataOffset;
  unsigned int *puiData;
  vifpacket_t vpkt;

  iSPS2Device=_sps2Open();
  pvEERegisters=_sps2MapEERegisters(0,iSPS2Device);
  printf("%i\n",errno);
  pvVUMemory=_sps2MapVUMemory(0,iSPS2Device);

  printf("Device = %i, EE Registers = %p, VU Memory = %p\n",
	 iSPS2Device, pvEERegisters, pvVUMemory);

  pUncached=sps2Allocate(4096, SPS2_MAP_BLOCK_4K | SPS2_MAP_UNCACHED,
			 iSPS2Device);

  // I could just put 0 for the map flags since cached is default.
  pCached=sps2Remap(pUncached, SPS2_MAP_CACHED, iSPS2Device);

  /**
   * Step 1.
   * Create dma packet to send using purely uncached memory.  Do not flush
   */
  // init vif packet
  vifpacket_init(&vpkt, pUncached->pvStart);
  
  // add "end" tag
  vifpacket_end(&vpkt, 0);
  
  // add "unpack" to write at the start of vu1mem
  vifpacket_openupkcode( &vpkt, 0, V4_32, 4, 4 );

  ulDataOffset=(unsigned long) vpkt.current-(unsigned long) vpkt.base;

  // add data
  vifpacket_adddata(&vpkt, 0x00000000);
  vifpacket_adddata(&vpkt, 0x00000000);
  vifpacket_adddata(&vpkt, 0x00000000);
  vifpacket_adddata(&vpkt, 0x00000000);
  
  // close "unpack"
  vifpacket_closeupkcode(&vpkt);
  
  vifpacket_terminate(&vpkt);

  // Write out real data, uncached
  puiData=ulDataOffset+pUncached->pvStart;
  *puiData++=0x12345678;
  *puiData++=0x23456789;
  *puiData++=0x34567890;
  *puiData++=0x45678901;

  *EE_D1_TADR_OFF(pvEERegisters)=sps2GetPhysicalAddress(vpkt.base, pUncached);
  *EE_D1_CHCR_OFF(pvEERegisters)= 1 | (1<<2) | (1<<8);
  
  sps2WaitForDMA(1, iSPS2Device);


  /**
   * Step 2.
   * Create tag using uncached memory, but insert data using cached memory and
   * flush half way.  We expect the unflushed cached data will not be written
   * and instead the same data from the previous, uncached run.
   */
  // init vif packet
  vifpacket_init(&vpkt, pUncached->pvStart);
  
  // add "end" tag
  vifpacket_end(&vpkt, 0);
  
  // add "unpack" to write at the start of vu1mem
  vifpacket_openupkcode( &vpkt, 1, V4_32, 4, 4 );

  ulDataOffset=(unsigned long) vpkt.current-(unsigned long) vpkt.base;

  // add data
  vifpacket_adddata(&vpkt, 0x00000000);
  vifpacket_adddata(&vpkt, 0x00000000);
  vifpacket_adddata(&vpkt, 0x00000000);
  vifpacket_adddata(&vpkt, 0x00000000);
  
  // close "unpack"
  vifpacket_closeupkcode(&vpkt);
  
  vifpacket_terminate(&vpkt);

  sps2FlushCache(iSPS2Device);

  // Write out real data, cached
  puiData=ulDataOffset+pCached->pvStart;
  *puiData++=0x11111111;
  *puiData++=0x22222222;
  sps2FlushCache(iSPS2Device);
  *puiData++=0x33333333;
  *puiData++=0x44444444;

  *EE_D1_TADR_OFF(pvEERegisters)=sps2GetPhysicalAddress(vpkt.base, pUncached);
  *EE_D1_CHCR_OFF(pvEERegisters)= 1 | (1<<2) | (1<<8);
  
  sps2WaitForDMA(1, iSPS2Device);

  /**
   * Step 3.
   * Create tag in uncached memory as in step 2.
   * Create data in cached memory, again as in step 2, but flush cache
   * before transfer.
   */
  sps2FlushCache(iSPS2Device);
  // init vif packet
  vifpacket_init(&vpkt, pUncached->pvStart);
  
  // add "end" tag
  vifpacket_end(&vpkt, 0);
  
  // add "unpack" to write at the start of vu1mem
  vifpacket_openupkcode( &vpkt, 2, V4_32, 4, 4 );

  ulDataOffset=(unsigned long) vpkt.current-(unsigned long) vpkt.base;

  // add data
  vifpacket_adddata(&vpkt, 0x00000000);
  vifpacket_adddata(&vpkt, 0x00000000);
  vifpacket_adddata(&vpkt, 0x00000000);
  vifpacket_adddata(&vpkt, 0x00000000);
  
  // close "unpack"
  vifpacket_closeupkcode(&vpkt);
  
  vifpacket_terminate(&vpkt);

  sps2FlushCache(iSPS2Device);

  // Write out real data, cached
  puiData=ulDataOffset+pCached->pvStart;
  *puiData++=0x99999999;
  *puiData++=0x88888888;
  *puiData++=0x77777777;
  *puiData++=0x66666666;

  sps2FlushCache(iSPS2Device);

  *EE_D1_TADR_OFF(pvEERegisters)=sps2GetPhysicalAddress(vpkt.base, pUncached);
  *EE_D1_CHCR_OFF(pvEERegisters)= 1 | (1<<2) | (1<<8);
  
  sps2WaitForDMA(1, iSPS2Device);

  printf("Beginning of vu memory is\n\t%08x %08x %08x %08x\n\t%08x %08x %08x %08x\n\t%08x %08x %08x %08x\n",
	 VU1_MEM_OFF(pvVUMemory)[ 0],
	 VU1_MEM_OFF(pvVUMemory)[ 1],
	 VU1_MEM_OFF(pvVUMemory)[ 2],
	 VU1_MEM_OFF(pvVUMemory)[ 3],
	 VU1_MEM_OFF(pvVUMemory)[ 4],
	 VU1_MEM_OFF(pvVUMemory)[ 5],
	 VU1_MEM_OFF(pvVUMemory)[ 6],
	 VU1_MEM_OFF(pvVUMemory)[ 7],
	 VU1_MEM_OFF(pvVUMemory)[ 8],
	 VU1_MEM_OFF(pvVUMemory)[ 9],
	 VU1_MEM_OFF(pvVUMemory)[10],
	 VU1_MEM_OFF(pvVUMemory)[11]);

  return 0;
}
